-module(eTpf).

-include("eTpf.hrl").

-export([
	%% 记录函数
	sts/0             %% tpTracerShell
	, stl/0           %% tpTracerLog
	, stf/0           %% tpTracerFile


	%% pt函数
	, pts/0           %% tpTracerShell
	, ptl/0           %% tpTracerLog
	, ptf/0           %% tpTracerFile

	%% pf函数
	, rsPfs/1
	, rsPfm/2

	, sdPfs/2
	, sdPfm/3

	, flPfs/2
	, flPfm/3

	, cgPfs/2
	, cgPfs/3
	, cgPfm/3
	, cgPfm/4

	%% trace函数
	, trace/1
	, trace/2
	, trace/3
	, trace/4
	, stop/0
	, stop/1
]).

sts() ->
	trace([{app, eSync}, {scope, [whereis(esSyncSrv)]}]).

stl() ->
	trace([{app, eSync}, {scope, [new_processes]}], tpTracerLog).

stf() ->
	trace([{app, eSync}, {scope, [whereis(esSyncSrv)]}], tpTracerFile).

pts() ->
	ok.

ptl() ->
	ok.


ptf() ->
	ok.

rsPfs(InputFile) ->
	tpMsgRS:pfs(InputFile).

rsPfm(InputFiles, Cwd) ->
	tpMsgRS:pfm(InputFiles, Cwd).

sdPfs(InputFile, Pids) ->
	tpMsgSD:pfs(InputFile, Pids).

sdPfm(InputFiles, Cwd, Pids) ->
	tpMsgSD:pfm(InputFiles, Cwd, Pids).

flPfs(InputFile, OutputPath) ->
	tpFlame:pfs(InputFile, OutputPath).

flPfm(InputFiles, Cwd, OutputPath) ->
	tpFlame:pfm(InputFiles, Cwd, OutputPath).

cgPfs(InputFile, Output) ->
	tpCallGrind:pfs(InputFile, Output).

cgPfs(InputFile, Output, Opts) ->
	tpCallGrind:pfs(InputFile, Output, Opts).

cgPfm(Wildcard, Cwd, Prefix) ->
	tpCallGrind:pfm(Wildcard, Cwd, Prefix).

cgPfm(Wildcard, Cwd, Prefix, Opts) ->
	tpCallGrind:pfm(Wildcard, Cwd, Prefix, Opts).

-spec trace(userInput()) -> ok.
trace(Input) ->
	trace(Input, tpTracerShell).

-spec trace(userInput(), module()) -> ok.
trace(Input, TracerMod) ->
	trace(Input, TracerMod, #{}, #{}).

-spec trace(userInput(), module(), tracerOpts()) -> ok.
trace(Input, TracerMod, TracerOpts) ->
	trace(Input, TracerMod, TracerOpts, #{}).

-spec trace(userInput(), module(), tracerOpts(), traceOpts()) -> ok.
trace(Input, TracerMod, TracerOpts, TraceOpts) when is_list(Input) ->
	InputList = case is_list(Input) of true -> Input; _ -> [Input] end,
	doTrace(InputList, TracerMod, TracerOpts, TraceOpts).

stop() ->
	stop(?eTpfTracerId).

stop(TracerId) ->
	supervisor:terminate_child(eTpf_sup, TracerId).

doTrace(InputList, TracerMod, TracerOpts, TraceOpts) ->
	_ = application:ensure_all_started(eTpf),

	TracerId = maps:get(tracerId, TraceOpts, ?eTpfTracerId),
	case supervisor:start_child(eTpf_sup, #{
		id => TracerId,
		start => {TracerMod, start_link, [TracerOpts]},
		restart => temporary,
		type => worker,
		shutdown => infinity
	}) of
		{ok, TracerPid} ->
			_Mode = maps:get(mode, TraceOpts, trace),
			Tem0InputList = flattenInput(InputList, []),
			Tem1InputList = ensurePattern(Tem0InputList),
			LastInputList = ensureScope(Tem1InputList),
			traceInput(LastInputList, TracerPid, TraceOpts),
			ok;
		_Err ->
			io:format("trace start error ~p~n", [_Err])
	end.

flattenInput([], Acc) ->
	lists:flatten(Acc);
flattenInput([{callback, Mod, Fun} | Tail], Acc) when is_atom(Mod), is_atom(Fun) ->
	Input = flattenInput(Mod:Fun(), []),
	flattenInput(Tail, [Input | Acc]);
flattenInput([{app, App} | Tail], Acc) when is_atom(App) ->
	_ = application:load(App),
	{ok, Mods} = application:get_key(App, modules),
	flattenInput(Tail, [Mods | Acc]);
flattenInput([Input | Tail], Acc) ->
	flattenInput(Tail, [Input | Acc]).

ensurePattern(Input) ->
	case [S || S = {scope, _} <- Input] of
		Input -> ['_' | Input];
		_ -> Input
	end.

ensureScope(Input) ->
	case [S || S = {scope, _} <- Input] of
		[] -> [{scope, [processes]} | Input];
		_ -> Input
	end.

traceInput([], _, _) ->
	ok;
traceInput([{scope, Scope} | Tail], TracerState, TraceOpts) ->
	% 每个进程只能由一个tracer进行跟踪。因此，跟踪已跟踪进程会失败。

	% PidPortSpec
	%%   all All currently existing processes and ports and all that will be created in the future.
	%%   processes All currently existing processes and all that will be created in the future.
	%%   ports All currently existing ports and all that will be created in the future.
	%%   existing All currently existing processes and ports.
	%%   existing_processes All currently existing processes.
	%%   existing_ports All currently existing ports.
	%%   new All processes and ports that will be created in the future.
	%%   new_processes All processes that will be created in the future.
	%%   new_ports All ports that will be created in the future.

	% FlagList
	%% We currently enable the following trace flags:
	%% - call: function calls
	%% - procs: process exit events; plus others we ignore
	%% - running: process being scheduled in/out
	%% - timestamp: events include the current timestamp
	%% - arity: function calls only include the arity, not arguments
	%% - return_to: return from functions
	%% - set_on_spawn: propagate trace flags to any children processes
	%%
	%% @todo It might be useful to count the number of sends
	%% or receives a function does.
	%ExtraFlags = [running || maps:get(running, TraceOpts, false)] ++ [send || maps:get(send, TraceOpts, false)],
	[
		begin
		%erlang:trace(PidPortSpec, true, [call, procs, arity, return_to, set_on_spawn, {tracer, tpTracerNif, TracerState} | ExtraFlags])
			erlang:trace(PidPortSpec, true, [call, arity, send, 'receive', set_on_spawn, {tracer, tpTracerNif, TracerState}])
		end || PidPortSpec <- Scope
	],
	traceInput(Tail, TracerState, TraceOpts);
traceInput([Mod | Tail], TracerState, TraceOpts) when is_atom(Mod) ->
	MatchSpec =
		case TraceOpts of
			#{process_dump := true} ->
				[{'_', [], [{message, {process_dump}}]}];
			_ ->
				true
		end,
	%% The module must be loaded before we attempt to trace it.
	_ = code:ensure_loaded(Mod),
	io:format("IMY******************  ~p~n  ", [Mod]),
	_ = erlang:trace_pattern({Mod, '_', '_'}, MatchSpec, [local]),
	traceInput(Tail, TracerState, TraceOpts).
